#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "pbeditcd.h"

#ifdef __GLIBC__
#if __GLIBC__ >= 2
#define stricmp strcasecmp
#define strnicmp strncasecmp
#endif
#endif

#define MAX_COLORS		8
#define MAX_BUBBLES		88
#define MAX_LEVELS		31	// 30 levels + the demo level

#define MAX_LINE_SIZE	80

#define PRG_FILE		"PB_PROG.PRG"
#define PRG_LENGTH		0x00038db2

#define OFS_LEVELS		0x0001e8b8


/* Program actions - default = EA_UNKNOWN */
enum	{EA_UNKNOWN = 0, EA_LOADPRG, EA_SAVEPRG, EA_LOADLVL, EA_SAVELVL, EA_INFOS};

/* Text lines actions */
enum	{TL_IGNORE = 0, TL_LEVEL, TL_COLOR, TL_INFOS};

/* Bubbles colors */
enum	{BC_RANDOM = 0, BC_BLUE, BC_RED, BC_YELLOW, BC_GREEN, BC_PURPLE, BC_ORANGE, BC_BLACK, BC_WHITE, BC_EMPTY, BC_END};

char*	bc_string[BC_END + 1] =
		{"random", "blue", "red", "yellow", "green", "purple", "orange", "black", "white", "empty", 0};
char*	bc_short_string[BC_END + 1] =
		{"[??]", "[bl]", "[rd]", "[ye]", "[gn]", "[pu]", "[or]", "[bk]", "[wh]", "....", 0};

UINT32	levels_table;

UINT16	nb_levels;

char*	file_data;

UINT8	quiet_mode = 0;
UINT8	action = EA_UNKNOWN;

char*	s_level	= "Level ";
char*	s_start	= "Start address ";
char*	s_color	= "Random colors ";

struct	level
{
	UINT8	number;
	UINT32	start;
	UINT8	nb_rndcolors;
	UINT8	tb_rndcolors[MAX_COLORS];
	UINT8	data[MAX_BUBBLES];
};

static struct level tb_level[MAX_LEVELS];


void CLIB_DECL logerror(const char *text,...)
{
	va_list arg;
	va_start(arg,text);
	if (!quiet_mode)
		vfprintf(stderr,text,arg);
	va_end(arg);
	return;
}


int load_file(const char *name, int length, char *data)
{
	FILE	*f;
	int	b_read;

	f = fopen(name,"rb");
	if (!f)
	{
		logerror("\nERROR ! File '%s' not found !",name);
		return 0;
	}
	else
	{
		fseek (f, 0L, SEEK_SET);
		b_read = fread(data, 1, length, f);
		fclose(f);
		return b_read;
	}
}

int save_file(const char *name, int length, char *data)
{
	FILE	*f;
	int	b_write;

	f = fopen(name,"wb");
	if (!f)
	{
		logerror("\nERROR ! File '%s' not found or read protected !",name);
		return 0;
	}
	else
	{
		fseek (f, 0L, SEEK_SET);
		b_write = fwrite(data, 1, length, f);
		fclose(f);
		return b_write;
	}
}

int load_prg(void)
{
	int	b_read;

	b_read = load_file(PRG_FILE,PRG_LENGTH,file_data);
	if (b_read == PRG_LENGTH)
	{
		levels_table = OFS_LEVELS + 4;
		nb_levels = MAX_LEVELS - 1;
		logerror("0x%08X bytes read.\n",b_read);
		logerror("Levels table stored at address 0x%08X.\n",levels_table);
		logerror("%d levels found.\n",nb_levels);
		return 1;
	}
	else
	{
		if (b_read)
			logerror("\nERROR ! Only 0x%08X bytes were read !\n",b_read);
		return 0;
	}
}

int save_prg(void)
{
	int	b_write;

	logerror("Writing PRG file ... ");
	b_write = save_file(PRG_FILE,PRG_LENGTH,file_data);
	if (b_write == PRG_LENGTH)
	{
		logerror("0x%08X bytes written.\n",b_write);
		return 1;
	}
	else
	{
		if (b_write)
			logerror("\nERROR ! Only 0x%08X bytes were written !\n",b_write);
		return 0;
	}
}

int match_bc_string(char *bc)
{
	int	i;

	for (i = 0;i < BC_END;i++)
		if (!strncmp(bc,bc_string[i],strlen(bc_string[i])))
			break;

	return i;
}

int match_bc_short_string(char *bc)
{
	int	i;

	for (i = 0;i < BC_END;i++)
		if (!strncmp(bc,bc_short_string[i],4))
			break;

	return i;
}

void data_to_tb_level(void)
{
	int		i,j;
	UINT32	offset_table;
	UINT32	offset_level;

	logerror("Copying PRG data to structure ... ");

	for (i = 1;(i <= nb_levels) && (i < MAX_LEVELS);i++)
	{
		offset_table = levels_table + 4 * (i - 1);
		offset_level = R_DWORD(file_data,offset_table);

		tb_level[i].number = i;
		tb_level[i].start = offset_level;
		tb_level[i].nb_rndcolors = R_BYTE(file_data,offset_level + 1);
		offset_level += 2;

		for (j = 0;j < tb_level[i].nb_rndcolors;j++)
		{
			tb_level[i].tb_rndcolors[j] = R_BYTE(file_data,offset_level + 2 * j + 1) / 2;
		}
		for (j = tb_level[i].nb_rndcolors;j < MAX_COLORS;j++)
			tb_level[i].tb_rndcolors[j] = 0;
		offset_level += 2 * tb_level[i].nb_rndcolors;

		for (j = 0;j < MAX_BUBBLES;j++)
		{
			switch R_BYTE(file_data,offset_level + 2 * j + 1)
			{
				case 0xfe:
					tb_level[i].data[j] = (UINT8)BC_EMPTY;
					break;
				case 0xff:
					tb_level[i].data[j] = (UINT8)BC_END;
					break;
				default:
					tb_level[i].data[j] = R_BYTE(file_data,offset_level + 2 * j + 1) / 2;
					break;
			}
		}
		offset_level += 2 * MAX_BUBBLES;
	}

	logerror("\n");

	return;
}

int text_to_tb_level(const char *name)
{
	FILE	*f;
	char  text_line[MAX_LINE_SIZE];
	int	i,j,p;
	int	tl_action;
	int	level_number = 1;		/* Range : 1 to MAX_LEVELS  - 1 */
	int	line_number = 1;		/* Range : 1 to MAX_BUBBLES / 8 */
	int	bubble_number = 0;		/* Range : 0 to MAX_BUBBLES - 1 */
	int	color_number = 0;		/* Range : 0 to MAX_COLORS  - 1 */

	logerror("Copying text file to structure ... ");

	f = fopen(name,"rb");
	if (!f)
	{
		logerror("\nERROR ! Levels text file '%s' not found !\n",name);
		return 0;
	}
	else
	{
		for (i = 1;(i <= nb_levels) && (i < MAX_LEVELS);i++)
		{
			tb_level[i].nb_rndcolors = 0;
			for (j = 0;j < MAX_COLORS;j++)
				tb_level[i].tb_rndcolors[j] = 0;
		}

		while (!feof(f) && (level_number <= nb_levels) && (level_number < MAX_LEVELS))
		{
			fgets(text_line,MAX_LINE_SIZE,f);
			tl_action = TL_INFOS;

			if ((!text_line) || (!text_line[0]))			/* Empty lines */
				tl_action = TL_IGNORE;
			if ((text_line[0] == '\n') || (text_line[0] == '\r'))	/* Empty lines */
				tl_action = TL_IGNORE;
			if ((text_line[0] == ';') || (text_line[0] == '#'))	/* Comments lines */
				tl_action = TL_IGNORE;

			if (!strncmp(text_line,s_level,strlen(s_level)))
				tl_action = TL_LEVEL;

			if (!strncmp(text_line,s_color,strlen(s_color)))
				tl_action = TL_COLOR;

			if ((text_line[0] != '|') && (tl_action == TL_INFOS))
				tl_action = TL_IGNORE;

			switch (tl_action)
			{
				case TL_LEVEL:
					break;
				case TL_COLOR:
					p = strlen(s_color) + 2;
					while ((p < strlen(text_line)) && (color_number < MAX_COLORS))
					{
						j = match_bc_string(&text_line[p]);
						if (j <= MAX_COLORS)
						{
							tb_level[level_number].nb_rndcolors = (UINT8)(color_number + 1);
							tb_level[level_number].tb_rndcolors[color_number] = (UINT8)j;
							color_number++;
							p += strlen(bc_string[j]) + 2;
						}
						else
						{
							p++;
							while ((p < strlen(text_line)) && (text_line[p - 1] != ' '))
								p++;
						}
					}
					break;
				case TL_INFOS:
					if (line_number % 2)				/* Odd lines */
					{
						p = 1;
						for (i = 0;i < 8;i++)
						{
							j = match_bc_short_string(&text_line[p]);
							/* If color not found, I put an empty space */
							if (j == BC_END)
								j = BC_EMPTY;
							/* If random color and 0 random colors, I put an empty space */
							if ((j == BC_RANDOM) &&	(!tb_level[level_number].nb_rndcolors))
								j = BC_EMPTY;
							tb_level[level_number].data[bubble_number] = (UINT8)j;
							bubble_number++;
							p += 4;
						}
					}
					else							/* Even lines */
					{
						p = 3;
						for (i = 0;i < 7;i++)
						{
							j = match_bc_short_string(&text_line[p]);
							/* If color not found, I put an empty space */
							if (j == BC_END)
								j = BC_EMPTY;
							/* If random color and 0 random colors, I put an empty space */
							if ((j == BC_RANDOM) && (!tb_level[level_number].nb_rndcolors))
								j = BC_EMPTY;
							tb_level[level_number].data[bubble_number] = (UINT8)j;
							bubble_number++;
							p += 4;
						}
						tb_level[level_number].data[bubble_number] = (UINT8)BC_END;
						bubble_number++;
					}

					line_number++;
					if (line_number > MAX_BUBBLES / 8)
					{
						level_number++;
						line_number = 1;
						bubble_number = 0;
						color_number = 0;
					}

					break;
			}
		}
	}

	logerror("\n");

	return 1;
}

void tb_level_to_data(void)
{
	int		i,j;
	UINT32	offset_table;
	UINT32	offset_level;

	logerror("Copying structure to PRG file ... ");

	for (i = 1;(i <= nb_levels) && (i < MAX_LEVELS);i++)
	{
		offset_level = tb_level[i].start;

		W_WORD(file_data,offset_level,tb_level[i].nb_rndcolors);
		offset_level += 2;

		for (j = 0;j < tb_level[i].nb_rndcolors;j++)
			W_WORD(file_data,offset_level + 2 * j,tb_level[i].tb_rndcolors[j] * 2);
		offset_level += 2 * tb_level[i].nb_rndcolors;

		for (j = 0;j < MAX_BUBBLES;j++)
		{
			switch (tb_level[i].data[j])
			{
				case (UINT8)BC_EMPTY:
					W_WORD(file_data,offset_level + 2 * j,0xfffe);
					break;
				case (UINT8)BC_END:
					W_WORD(file_data,offset_level + 2 * j,0xffff);
					break;
				default:
					W_WORD(file_data,offset_level + 2 * j,tb_level[i].data[j] * 2);
					break;
			}
		}
		offset_level += 2 * MAX_BUBBLES;
	}

	logerror("\n");

	return;
}

void display_one_level(FILE *f_output, int level_number)
{
	int	j;

	fprintf(f_output,"%s%2d       (%s: 0x%08x)\n", \
			s_level,tb_level[level_number].number,s_start,tb_level[level_number].start);

	if (tb_level[level_number].nb_rndcolors)
	{
		fprintf(f_output,"%s: ", s_color);
		fprintf(f_output,"%s",bc_string[tb_level[level_number].tb_rndcolors[0]]);
		for (j = 1;j < tb_level[level_number].nb_rndcolors;j++)
			fprintf(f_output,", %s", bc_string[tb_level[level_number].tb_rndcolors[j]]);
		fprintf(f_output,".\n");
	}
	fprintf(f_output,"\n");

	for (j = 0;j < MAX_BUBBLES;j++)
	{
		if (j % 8 == 0)
			fprintf(f_output,"|");
		if (j % 16 == 8)
			fprintf(f_output,"..");
		if (tb_level[level_number].data[j] != BC_END)
			fprintf(f_output,"%s", bc_short_string[tb_level[level_number].data[j]]);
		if (j % 16 == 15)
			fprintf(f_output,"..");
		if (j % 8 == 7)
			fprintf(f_output,"|\n");
	}
	fprintf(f_output,"\n\n");

	return;
}

void display_levels(FILE *f_output)
{
	int	i;

	fprintf(f_output,"# NEOCD Puzzle Bobble level editor - version %s\n# Copyright %s, %s\n\n\n", \
			VERSION,LAST_CODER,LAST_UPDATE);

	for (i = 1;(i <= nb_levels) && (i < MAX_LEVELS);i++)
	{
		display_one_level(f_output,i);
	}
}

void run_action(int action)
{
	FILE  *f_output;
	char	name[256];

	sprintf (name, "%s.lvl", UTL_FILE);

	switch (action)
	{
		case EA_LOADPRG:
			data_to_tb_level();
			display_levels(stdout);
			break;
		case EA_SAVEPRG:
			data_to_tb_level();	/* To fill all fields that aren't in the text file */
			text_to_tb_level(name);
			tb_level_to_data();
			save_prg();
			break;
		case EA_LOADLVL:
			data_to_tb_level();	/* To fill all fields that aren't in the text file */
			if (text_to_tb_level(name))
				display_levels(stdout);
			break;
		case EA_SAVELVL:
			data_to_tb_level();
			logerror("Creating levels text file ... ");
			f_output = fopen(name,"w");
			display_levels(f_output);
			fclose(f_output);
			logerror("'%s' generated.\n",name);
			break;
		case EA_INFOS:
			data_to_tb_level();
			break;
	}
	return;
}

void show_help(void)
{
	logerror("NEOCD Puzzle Bobble level editor - version %s\nCopyright %s, %s\n\n", \
			VERSION,LAST_CODER,LAST_UPDATE);
	logerror("Syntax : %s <action>\n",UTL_NAME);
	logerror("  <action>   = '-loadprg' (displays levels from PRG file)\n");
	logerror("               '-loadlvl' (displays levels from text file)\n");
	logerror("               '-saveprg' (saves levels from text file to PRG file)\n");
	logerror("               '-savelvl' (saves levels from PRG file to text file)\n");
	logerror("\n");
	return;
}

void run_editor(int action, int argc, char **argv)
{
	if (action == EA_UNKNOWN)
	{
		show_help();
		return;
	}

	logerror("Reading PRG file ... ");
	if (load_prg())
	{
		run_action(action);
	}
	return;
}

int main (int argc, char **argv)
{
	int	i;

	for (i = 1;i < argc;i++)
		if (argv[i][0] == '/')
			argv[i][0] = '-';

	for (i = 1;i < argc;i++)
	{
		if (!stricmp(argv[i],"-loadprg"))	action = EA_LOADPRG;
		if (!stricmp(argv[i],"-saveprg"))	action = EA_SAVEPRG;
		if (!stricmp(argv[i],"-loadlvl"))	action = EA_LOADLVL;
		if (!stricmp(argv[i],"-savelvl"))	action = EA_SAVELVL;

		if (!stricmp(argv[i],"-infos"))	action = EA_INFOS;

		if (!stricmp(argv[i],"-quiet"))	quiet_mode = 1;
	}

	file_data = (char*)malloc(PRG_LENGTH);
	if (file_data)
	{
		run_editor(action,argc,argv);
		free(file_data);
	}
	else
		logerror("ERROR ! Not enough memory to load the file(s) !\n");

	return 0;
}

